"""
The zenodo url version is stored here. Makes it easier to update when new files are uploaded.

TODO: future version may hold the file names here too.
"""


zenodo_url = "https://zenodo.org/record/4664653"

year_range = [1985, 2019]
version = "v2021.02"
contact = "luke.gregor@usys.ethz.ch"
citation = (
    "Fay, A. R., Gregor, L., Landschuetzer, P., McKinley, G. A., Gruber, N., "
    "Gehlen, M., Iida, Y., Laruelle, G. G., Roedenbeck, C., and Zeng, J.: "
    "Harmonization of global surface ocean pCO2 mapped products and their "
    "flux calculations; an improved estimate of the ocean carbon sink, Earth "
    "Syst. Sci. Data Discuss. [preprint], https://doi.org/10.5194/essd-2021-16, "
    "in review, 2021."
)
